/*
 * ProFTPD: mod_otp -- a module for authenticating via use of one-time
 *                     passwords (OTP).
 *
 * Copyright (c) 2002-2006 TJ Saunders
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * As a special exemption, TJ Saunders and other respective copyright holders
 * give permission to link this program with OpenSSL, and distribute the
 * resulting executable, without including the source code for OpenSSL in the
 * source distribution.
 *
 * This is mod_otp, contrib software for proftpd 1.2 and above.
 * For more information contact TJ Saunders <tj@castaglia.org>.
 *
 * $Id: mod_otp.c,v 1.3 2006/12/06 16:12:53 tj Exp tj $
 */

#include "conf.h"
#include "privs.h"

#include "mod_otp.h"

static unsigned char otp_engine = FALSE;

#if defined(USE_HEIMDAL_OTP)
#include <otp.h>
static OtpContext otp_ctx;

#elif defined(USE_OPIE)
#include <opie.h>
static struct opie otp_ctx;

#elif defined(USE_SKEY)
#include <skey.h>
static struct skey otp_ctx;
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t dst_sz) {
  size_t n;

  for (n = 0; n < dst_sz; n++) {
    if ((*dst++ = *src++) == '\0')
      break;
  }

  if (n < dst_sz)
    return n;

  if (n > 0)
    *(dst - 1) = '\0';

  return n + strlen(src);
}
#endif

/* Configuration handlers
 */

MODRET set_otpengine(cmd_rec *cmd) {
  int bool = -1;
  config_rec *c = NULL;

  CHECK_ARGS(cmd, 1);
  CHECK_CONF(cmd, CONF_ROOT|CONF_VIRTUAL|CONF_GLOBAL);

  bool = get_boolean(cmd, 1);
  if ((bool == -1)
    CONF_ERROR(cmd, "expecting Boolean parameter");

  c = add_config_param(cmd->argv[0], 1, NULL);
  c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
  *((unsigned char *) c->argv[0]) = bool;

  return HANDLED(cmd);
}

/* Command handlers
 */

MODRET otp_post_user(cmd_rec *cmd) {
  int res = -1;

#if defined(USE_OPIE)
  char challenge[OPIE_CHALLENGE_MAX+1];
#else
  char challenge[PR_TUNABLE_BUFFER_SIZE];
#endif

  if (!otp_engine)
    return DECLINED(cmd);

  memset(challenge, '\0', sizeof(challenge));

  PRIVS_ROOT
#if defined(USE_HEIMDAL_OTP)
  res = otp_challenge(&otp_ctx, cmd->argv[1], challenge, sizeof(challenge));
#elif defined(USE_OPIE)
  res = opiechallenge(&otp_ctx, cmd->argv[1], challenge);
#elif defined(USE_SKEY)
  res = skeychallenge(&otp_ctx, cmd->argv[1], challenge);
#endif
  PRIVS_RELINQUISH

  if (res) {
    char *str = NULL;

#if defined(USE_HEIMDAL_OTP)
    if ((str = otp_error(&otp_ctx)) != NULL)
#else
    if ((str = "(unknown)"))
#endif
      pr_log_debug(DEBUG3, MOD_OTP_VERSION
        ": error generating challenge for user '%s': %s", cmd->argv[1], str);

  } else
    pr_add_response_dup(R_DUP, "Challenge: [ %s ]", challenge);

  return DECLINED(cmd);
}

/* Authentication handlers
 */

MODRET otp_auth(cmd_rec *cmd) {
  register unsigned int i = 0;
  int res = -1;

#if defined(USE_OPIE)
  char response[OPIE_RESPONSE_MAX+1];
#else
  char response[PR_TUNABLE_BUFFER_SIZE];
#endif

  if (!otp_engine)
    return DECLINED(cmd);

  memset(response, '\0', sizeof(response));

  for (i = 0; i < sizeof(response) && i < strlen(cmd->argv[1]); i++)
    response[i] = toupper(cmd->argv[1][i]);

  PRIVS_ROOT
#if defined(USE_HEIMDAL_OTP)
  res = otp_verify_user(&otp_ctx, response);
#elif defined(USE_OPIE)
  res = opieverify(&otp_ctx, response);
#elif defined(USE_SKEY)
  res = skeyverify(&otp_ctx, response);
#endif
  PRIVS_RELINQUISH

  if (!res) {
    pr_log_auth(PR_LOG_NOTICE, MOD_OTP_VERSION ": user '%s' authenticated via "
      "one-time password (%s)", cmd->argv[0],
#if defined(USE_HEIMDAL_OPT)
      "Heimdal OTP"
#elif defined(USE_OPIE)
      "OPIE"
#elif defined(USE_SKEY)
      "S/Key"
#endif
      );
    return HANDLED(cmd);
  }

  return DECLINED(cmd);
}

/* Initialization functions
 */

static int otp_sess_init(void) {
  unsigned char *use_otp = NULL;

  use_otp = get_param_ptr(main_server->conf, "OTPEngine", FALSE);
  if (use_otp != NULL)
    otp_engine = *use_otp;

  return 0;
}

/* Module API tables
 */

static conftable otp_conftab[] = {
  { "OTPEngine",	set_otpengine,	NULL },
  { NULL }
};

static cmdtable otp_cmdtab[] = {
  { POST_CMD,	C_USER,	G_NONE,	otp_post_user,		FALSE,	FALSE },
  { 0, NULL }
};

static authtable otp_authtab[] = {
  { 0, "auth",	otp_auth },
  { 0, NULL }
};

module otp_module = {
  NULL, NULL,

  /* Module API version 2.0 */
  0x20,

  /* Module name */
  "otp",

  /* Module configuration handler table */
  otp_conftab,

  /* Module command handler table */
  otp_cmdtab,

  /* Module authentication handler table */
  otp_authtab,

  /* Module initialization function */
  NULL,

  /* Session initialization function */
  otp_sess_init
};
